"""`bravais.validator.py`"""

import copy
import random


class ValidationError(Exception):
    """Raised when NumericValidator.validate fails."""


class NumericValidator(object):
    """Class to represent a set of validation conditions to be applied to
    a set of numeric parameters."""

    def __init__(self, equal_to=None, not_equal_to=None, equal_groups=None,
                 not_equal_groups=None, default_min=0, default_max=1):
        """
        Initialise a NumericValidator with a set of conditions.

        Parameters
        ----------
        equal_to : dict, optional
        not_equal_to : dict, optional
        equal_groups : list of list of str, optional
        not_equal_groups : list of list of str, optional
        default_min : int or float, optional
        default_max : int or float, optional

        """

        if equal_to is not None:
            if not isinstance(equal_to, dict):
                msg = '`equal_to` must be a dict.'
                raise ValueError(msg)

        if not_equal_to is not None:
            if not isinstance(not_equal_to, dict):
                msg = '`not_equal_to` must be a dict.'

        if equal_groups is not None:
            msg = '`equal_groups` must be a list of list of str.'
            if isinstance(equal_groups, list):
                for i in equal_groups:
                    if isinstance(i, list):
                        for j in i:
                            if not isinstance(j, str):
                                raise ValueError(msg)
                    else:
                        raise ValueError(msg)
            else:
                raise ValueError(msg)

        if not_equal_groups is not None:
            msg = '`not_equal_groups` must be a list of list of str.'
            if isinstance(not_equal_groups, list):
                for i in not_equal_groups:
                    if isinstance(i, list):
                        for j in i:
                            if not isinstance(j, str):
                                raise ValueError(msg)
                    else:
                        raise ValueError(msg)
            else:
                raise ValueError(msg)

        self.equal_to = equal_to or {}
        self.not_equal_to = not_equal_to or {}
        self.equal_groups = equal_groups or [[]]
        self.not_equal_groups = not_equal_groups or [[]]
        self.default_min = default_min
        self.default_max = default_max

        self._validate_conditions()

    def _get_parameter_equal_group(self, param):
        """Get the `equal_group` in which a parameter exists, if at all."""
        for i in self.equal_groups:
            if param in i:
                return i

    def _get_parameter_not_equal_group(self, param):
        """Get the `not_equal_group` in which a parameter exists, if at all."""
        for i in self.not_equal_groups:
            if param in i:
                return i

    def _validate_conditions(self):
        """Check conditions make sense."""

        msg = ('Parameter names should not be repeated within equal_groups` '
               'elements, nor within `not_equal_groups` elements')
        for i in (self.equal_groups + self.not_equal_groups):
            if len(set(i)) != len(i):
                raise ValueError(msg)

        msg = ('The parameter "{}" appears in more than one sublist of '
               '`equal_groups`; these two sublists should be merged.')
        eq_groups_params = []
        for i in self.equal_groups:
            for j in i:
                if j in eq_groups_params:
                    raise ValueError(msg.format(j))
                eq_groups_params.append(j)

        msg = ('The parameter "{}" appears in more than one sublist of '
               '`not_equal_groups`; these two sublist should be merged.')
        neq_groups_params = []
        for i in self.not_equal_groups:
            for j in i:
                if j in neq_groups_params:
                    raise ValueError(msg.format(j))
                neq_groups_params.append(j)

        msg = ('The parameters "{}" are specified in both `equal_to` and '
               '`not_equal_to`. Specify the parameter in only one of these.')
        eq_to_params = self.equal_to.keys()
        neq_to_params = self.not_equal_to.keys()
        intersect = list(set(eq_to_params) & set(neq_to_params))
        if intersect:
            raise ValueError(msg.format(intersect))

        msg = ('Only one of the parameters specified within an `equal_group`'
               'element may be specified in `equal_to`.')
        for i in self.equal_groups:
            if len(list(set(i) & set(self.equal_to.keys()))) > 1:
                raise ValueError(msg)

        msg = ('Parameters specified within a `not_equal_group` element cannot'
               ' be assigned to the same value in `equal_to`.')
        for i in self.not_equal_groups:
            eq_to_vals = [self.equal_to.get(j) for j in i
                          if self.equal_to.get(j) is not None]
            if len(set(eq_to_vals)) < len(eq_to_vals):
                raise ValueError(msg)

    def _set_default_value(self, all_params, params_to_set):
        """Assign a default value to a parameter.

        Parameters
        ----------
        all_params : dict
            All parameters. The the value of the parameter specified will be
            changed within this dict.
        params_to_set : list of str
            Names of the parameters to be set to the same value.

        Returns
        -------
        None

        """

        assert all([i in all_params for i in params_to_set])

        # Collect values param must not be:
        bad_vals = []
        for i in params_to_set:
            if i in self.not_equal_to:
                bad_vals.append(self.not_equal_to[i])

            ne_group = self._get_parameter_not_equal_group(i) or []
            for j in ne_group:
                if j != i:
                    ne_val = all_params[j]
                    if ne_val:
                        bad_vals.append(ne_val)

        # Find a suitable default value:
        default_val = None
        count = 0
        while default_val is None or default_val in bad_vals:

            default_val = random.random()
            default_val *= (self.default_max - self.default_min)
            default_val += self.default_min

            count += 1
            if count > 10:
                msg = ('Failed to find suitable default value for '
                       'parameter {}.')
                raise RuntimeError(msg.format(params_to_set))

        for i in params_to_set:
            all_params[i] = default_val

    def validate(self, **kwargs):
        """Validate named parameters."""

        params = kwargs.copy()

        # Firstly, validate according to rules, but ignore parameters set to
        # `None`:

        msg = 'Parameter "{}" must be equal to "{}".'
        for param_name, val in self.equal_to.items():
            if kwargs[param_name] is not None and kwargs[param_name] != val:
                raise ValidationError(msg.format(param_name, val))

        msg = 'Parameter "{}" must not be equal to "{}".'
        for param_name, val in self.not_equal_to.items():
            if kwargs[param_name] is not None and kwargs[param_name] == val:
                raise ValidationError(msg.format(param_name, val))

        msg = 'Parameters "{}" must all have the same value.'
        for i in self.equal_groups:
            first_val = None
            for j in i:
                if first_val is None and kwargs[j] is not None:
                    first_val = kwargs[j]
                    continue
                if kwargs[j] is not None and kwargs[j] != first_val:
                    raise ValidationError(msg.format(i))

        msg = 'Parameters "{}" must all have distinct values.'
        for i in self.not_equal_groups:
            all_vals = []
            for j in i:
                if kwargs[j] is not None:
                    if kwargs[j] in all_vals:
                        raise ValidationError(msg.format(i))
                    all_vals.append(kwargs[j])

        none_params = [k for k, v in kwargs.items() if v is None]

        # Fill in `None`s.
        # First set any `equal_to`s:
        for i in copy.copy(none_params):
            if i in self.equal_to:
                params[i] = self.equal_to[i]
                none_params.remove(i)

        # Then try using `equal_groups`:
        for eq_group in self.equal_groups:

            num_none = sum([params.get(i) is None for i in eq_group])

            if len(eq_group) > num_none > 0:
                # At least one param in the eq_group has a value, use this one.

                val = None
                for i in eq_group:
                    if params[i] is not None:
                        val = params[i]
                        break

                for i in eq_group:
                    if params[i] is None:
                        params[i] = val
                        none_params.remove(i)

            elif len(eq_group) == num_none:
                # None of the equal group values are set. Use a default.
                self._set_default_value(params, eq_group)
                for i in eq_group:
                    none_params.remove(i)

        # Try to set remaining `None`s using defaults. Need a default value
        # that doesn't conflict with `not_equal_to`, nor with
        # `not_equal_groups`.
        if none_params:
            for i in copy.copy(none_params):
                self._set_default_value(params, [i])
                none_params.remove(i)

        return params

    def __repr__(self):
        return (
            '<{}('
            'equal_to={}, '
            'not_equal_to={}, '
            'equal_groups={}, '
            'not_equal_groups={}'
            ')>'.format(
                self.__class__.__name__,
                self.equal_to,
                self.not_equal_to,
                self.equal_groups,
                self.not_equal_groups,
            )
        )



"""`bravais.bravais.py`

Module defining a BravaisLattice class.

"""

from pathlib import Path
import pkgutil
from enum import Enum

import numpy as np
import yaml

from bravais.validator import NumericValidator

# Lattice sites (as column vectors) in fractional coordinates, for each
# centring type:
CENTRING_LATTICE_SITES = {
    'primitive': np.array([
        [0.0],
        [0.0],
        [0.0],
    ]),
    'base': np.array([
        [0.0, 0.5],
        [0.0, 0.5],
        [0.0, 0.0],
    ]),
    'body': np.array([
        [0.0, 0.5],
        [0.0, 0.5],
        [0.0, 0.5],
    ]),
    'face': np.array([
        [0.0, 0.5, 0.5, 0.0],
        [0.0, 0.5, 0.0, 0.5],
        [0.0, 0.0, 0.5, 0.5],
    ]),
    'rhombohedral': np.array([
        [0.0, 1/3, 2/3],
        [0.0, 2/3, 1/3],
        [0.0, 1/3, 2/3],
    ]),
}


class CentringType(Enum):
    """Class to represent the centring type."""

    primitive = 'P'
    base = 'B'
    body = 'I'
    face = 'F'
    rhombohedral = 'R'


class BravaisLattice(object):
    """Class to represent a Bravais lattice unit cell.

    Attributes
    ----------
    lattice_system : str
        Lattice system. One of "cubic", "rhombohedral", "orthorhombic",
        "tetragonal", "monoclinic", "triclinic", "hexagonal".
    centring_type : str
        Centring type also known as lattice type.
    a : float
        Lattice parameter, magnitude of the first unit cell edge vector.
    b : float
        Lattice parameter, magnitude of the second unit cell edge vector.
    c : float
        Lattice parameter, magnitude of the third unit cell edge vector.
    alpha : float
        Lattice parameter, angle in degrees between the second and third unit
        cell edge vectors.
    beta : float
        Lattice parameter, angle in degrees between the first and third unit
        cell edge vectors.
    gamma : float
        Lattice parameter, angle in degrees between the first and second unit
        cell edge vectors.
    unit_cell : ndarray of shape (3, 3)
        Array of column or row (depending on `row_or_column`) vectors defining
        the lattice unit cell.
    lat_sites_std: ndarray of shape (3, N)
        Array of column or row (depending on `row_or_column`) vectors defining
        the Cartesian positions of lattice sites within the lattice unit cell.
    lat_sites_frac: ndarray of shape (3, N)
        Array of column or row (depending on `row_or_column`) vectors defining
        the fractional positions of lattice sites within the lattice unit cell.
    row_or_column : str ("row" or "column")
        Specified whether `unit_cell`, `lat_sites_std` and `lat_sites_frac` are
        arrays of row or column vectors.

    Notes
    -----
    Lattice vectors are formed by aligning the crystallographic x-axis (i.e
    with magnitude `a`) along the Cartesian x-axis and aligning the
    crystallographic xy-plane (i.e. the vectors with magnitudes `a` and `b`)
    parallel to the Cartesian xy-plane.

    Conventional unit cells are generated and expected in the lattice parameter
    parameters. For instance the rhombohedral lattice system is represented
    with a rhombohedrally-centred hexagonal unit cell (the hexagonal axes/
    "setting"), rather than a primitive rhombohedral (cube stretched along its
    diagonal) cell.

    References
    ----------
    TODO: <Insert reference here to the 14 Bravais lattices in 3D.>

    Examples
    --------
    TODO: <Insert example here>

    TODO:
    -   Add primitive centring type to allowed centring types for rhombohedral
        lattice system.
    -   Regarding restrictions/validations on lattice types, need to allow
        restrictions to be "any n parameters must be..." rather than "the first
        two parameters must be". E.g. monoclinic: "two of the angle parameters
        must be 90 deg" rather than "parameters 3 and 5 must be 90 deg".
    -   Add align option ('ax' or 'cz').

    """

    def __init__(self, lattice_system=None, centring_type=None, a=None, b=None,
                 c=None, α=None, β=None, γ=None, alpha=None, beta=None,
                 gamma=None, degrees=True, alignment='ax',
                 row_or_column='column'):
        """Constructor method for BravaisLattice object.

        Parameters
        ----------
        lattice_system : str
            Lattice system is one of: cubic, rhombohedral, orthorhombic,
            tetragonal, monoclinic, triclinic, hexagonal.
        centring_type : str, optional
            The centring type of the lattice, also known as the lattice type,
            is one of P (primitive), B (base-centred), I (body-centred), F
            (face-centred) or R (rhombohedrally-centred). Not all centring
            types are compatible with all lattice systems. Default is None, in
            which case the rhombohedrally-centred centring type (for the
            rhombohedral lattice system) or primitive centring type (for all
            other lattice systems) will be chosen.
        a : float, optional
            Lattice parameter, magnitude of the first unit cell edge vector.
        b : float, optional
            Lattice parameter, magnitude of the second unit cell edge vector.
        c : float, optional
            Lattice parameter, magnitude of the third unit cell edge vector.
        α : float, optional
            Lattice parameter, angle in degrees between the second and third
            unit cell edge vectors. Specify `alpha` or `α`, but not both.
        β : float, optional
            Lattice parameter, angle in degrees between the first and third
            unit cell edge vectors. Specify `beta` or `β`, but not both.
        γ : float, optional
            Lattice parameter, angle in degrees between the first and second
            unit cell edge vectors. Specify `gamma` or `γ`, but not both.
        alpha : float, optional
            Lattice parameter, angle in degrees between the second and third
            unit cell edge vectors. Specify `alpha` or `α`, but not both.
        beta : float, optional
            Lattice parameter, angle in degrees between the first and third
            unit cell edge vectors. Specify `beta` or `β`, but not both.
        gamma : float, optional
            Lattice parameter, angle in degrees between the first and second
            unit cell edge vectors. Specify `gamma` or `γ`, but not both.
        row_or_column : str ("row" or "column"), optional
            Defines vector direction in array attributes. For instance, if
            "row", unit cell edge vectors are represented as row vectors in
            `unit_cell`. If "column", unit cell edge vectors are represented as
            column vectors in `unit_cell`. Default is "column".

        """

        lat_cent = self._validate_lattice_system(lattice_system, centring_type)
        self.lattice_system = lat_cent[0]
        self.centring_type = lat_cent[1]

        alpha, beta, gamma = self._normalise_angle_spec(
            alpha, beta, gamma, α, β, γ)

        lengths, angles = self._validate_params(a, b, c, alpha, beta, gamma)
        self.a = lengths['a']
        self.b = lengths['b']
        self.c = lengths['c']
        self.alpha = angles['alpha']
        self.beta = angles['beta']
        self.gamma = angles['gamma']

        self.row_or_column = row_or_column
        self.alignment = alignment
        self._unit_cell = self._compute_unit_cell(alignment)
        self.lattice_sites_frac = CENTRING_LATTICE_SITES[self.centring_type.name]

        # TODO: add a `degrees` boolean attribute for the angles.
        # TODO: mustn't be able to assign centring_type and lattice_system or
        # lengths/angles after init

    def _normalise_angle_spec(self, alpha, beta, gamma, α, β, γ):
        """Check angles are not specified as both spelled-out and greek
        symbols."""

        for i, j in zip([alpha, beta, gamma], [α, β, γ]):
            if i is not None and j is not None:
                msg = ('For each angle, specify either the spelled-out version'
                       ' (e.g. "alpha"), or the Greek letter version (e.g. '
                       '"α"), but not both!.')
                raise ValueError(msg)

        alpha = alpha or α
        beta = beta or β
        gamma = gamma or γ

        return alpha, beta, gamma

    def _compute_unit_cell(self, alignment):
        """Use the lattice parameters to form the unit cell with a particular
        alignment with the Cartesian axes.

        Parameters
        ----------
        alignment : str
            Alignment of the unit cell with repsect to the Cartesian axes, must
            be one either "ax" (align the `a` lattice vector with the x-axis,
            and the `b` lattice vector in the xy-plane) or "cz" (align the `c`
            lattice vector with the z-axis and the `b` lattice vector in the
            yz-plane).

        Returns
        -------
        unit_cell : ndarray of shape (3, 3)
            Array of column vectors representing the unit cell edge vectors.

        TODO: add some references for these equations. Consider doing one
        original alignment, and then rotating from that? Might want to support
        arbitrary orientations, which would involve specifying the orientation
        (as a vector) of one of the lattice vectors, plus a plane (specified
        as a normal vector?) in which a second lattice vector should lie.

        """

        alignment_opts = ['ax', 'cz']
        if alignment not in alignment_opts:
            msg = ('"{}" is not a valid axes alignment option. `align` must be'
                   ' one of: {}.'.format(alignment, alignment_opts))
            raise ValueError(msg)

        alpha_r = np.deg2rad(self.alpha)
        beta_r = np.deg2rad(self.beta)
        gamma_r = np.deg2rad(self.gamma)

        if alignment == 'ax':
            # Align `a` lattice vector with the x-axis and `b` lattice vector
            # in the xy-plane.

            a_x = self.a
            b_x = self.b * np.cos(gamma_r)
            b_y = self.b * np.sin(gamma_r)
            c_x = self.c * np.cos(beta_r)
            c_y = (abs(self.c) * abs(self.b) *
                   np.cos(alpha_r) - b_x * c_x) / b_y
            c_z = np.sqrt(self.c**2 - c_x**2 - c_y**2)

            unit_cell = np.array([
                [a_x, b_x, c_x],
                [0,  b_y, c_y],
                [0,  0, c_z]
            ])

        elif alignment == 'cz':
            # Align `c` lattice vector with the z-axis and `b` lattice vector
            # in the yz-plane.

            f = (1 - (np.cos(alpha_r))**2 - (np.cos(beta_r))**2 - (np.cos(gamma_r))**2
                 + 2 * np.cos(alpha_r) * np.cos(beta_r) * np.cos(gamma_r))**0.5
            a_x = self.a * f / np.sin(alpha_r)
            a_y = self.a * (np.cos(gamma_r) - np.cos(alpha_r)
                            * np.cos(beta_r)) / np.sin(alpha_r)
            a_z = self.a * np.cos(beta_r)
            b_y = self.b * np.sin(alpha_r)
            b_z = self.b * np.cos(alpha_r)
            c_z = self.c

            unit_cell = np.array([
                [a_x, 0, 0],
                [a_y, b_y, 0],
                [a_z, b_z, c_z]
            ])

        return unit_cell

    def _validate_lattice_system(self, lattice_system, centring_type):
        """Validate the specified lattice system and centring type."""

        if centring_type is None:
            if lattice_system == 'rhombohedral':
                centring_type = 'R'
            else:
                centring_type = 'P'

        # List valid Bravais lattice systems and centring types:
        all_lat = [
            'triclinic',
            'monoclinic',
            'orthorhombic',
            'tetragonal',
            'rhombohedral',
            'hexagonal',
            'cubic'
        ]
        centring_compat = {
            'P': list(set(all_lat) - set(['rhombohedral'])),
            'B': ['monoclinic', 'orthorhombic'],
            'I': ['orthorhombic', 'tetragonal'],
            'F': ['orthorhombic', 'cubic'],
            'R': ['rhombohedral'],
        }
        all_cent = list(centring_compat.keys())

        lattice_system = lattice_system.lower()
        centring_type = centring_type.upper()

        # Check specified lattice system and centring type are compatible:
        if lattice_system not in all_lat:
            msg = ('"{}" is not a valid lattice system. `lattice_system` must '
                   'be one of: {}.')
            raise ValueError(msg.format(lattice_system, all_lat))

        if centring_type not in all_cent:
            msg = ('"{}" is not a valid centering type. `centring_type` must '
                   'be one of {}.')
            raise ValueError(msg.format(centring_type, all_cent))

        if lattice_system not in centring_compat[centring_type]:
            msg = 'Lattice system {} and centring type {} are not compatible.'
            raise ValueError(msg.format(lattice_system, centring_type))

        centring_type = CentringType(centring_type)

        return lattice_system, centring_type

    def _validate_params(self, a, b, c, alpha, beta, gamma):
        """Validate lattice parameters against the specified lattice system."""

        data = pkgutil.get_data('bravais', 'valid_parameters.yml')
        valid_parameters = yaml.load(data)

        valid_lat_params = valid_parameters[self.lattice_system]
        length_validator = NumericValidator(**valid_lat_params['lengths'])
        angles_validator = NumericValidator(**valid_lat_params['angles'])

        lengths_valid = length_validator.validate(a=a, b=b, c=c)
        angles_valid = angles_validator.validate(
            alpha=alpha, beta=beta, gamma=gamma)

        return lengths_valid, angles_valid

    @property
    def row_or_column(self):
        return self._row_or_column

    @row_or_column.setter
    def row_or_column(self, row_or_column):
        if row_or_column not in ['row', 'column']:
            msg = ('`row_or_column` must be specified as a string, either '
                   '"row" or "column".')
            raise ValueError(msg)
        self._row_or_column = row_or_column

    @property
    def unit_cell(self):
        if self.row_or_column == 'column':
            return self._unit_cell
        elif self.row_or_column == 'row':
            return self._unit_cell.T

    @property
    def lattice_sites(self):
        """Get the position of the lattice sites in Cartesian coordinates."""
        return np.dot(self._unit_cell, self.lattice_sites_frac)

    @property
    def α(self):
        return self.alpha

    @property
    def β(self):
        return self.beta

    @property
    def γ(self):
        return self.gamma

    def __repr__(self):
        return (
            '<{}('
            'lattice_system="{}", '
            'centring_type="{}", '
            'a={:.4f}, '
            'b={:.4f}, '
            'c={:.4f}, '
            'alpha={:.2f}, '
            'beta={:.2f}, '
            'gamma={:.2f}'
            ')>'.format(
                self.__class__.__name__,
                self.lattice_system,
                self.centring_type.value,
                self.a,
                self.b,
                self.c,
                self.alpha,
                self.beta,
                self.gamma,
            )
        )

    def __str__(self):
        return (
            '{}-centred {} lattice ('
            'a={:.4f}, b={:.4f}, c={:.4f}, '
            'alpha={:.2f}, beta={:.2f}, gamma={:.2f}'
            ')'.format(
                self.centring_type.value,
                self.lattice_system,
                self.a,
                self.b,
                self.c,
                self.alpha,
                self.beta,
                self.gamma,
            )
        )



"""Pip installation script for `bravais`."""

import os
import re
from setuptools import find_packages, setup


def get_version():

    ver_file = '{}/_version.py'.format('bravais')
    with open(ver_file) as handle:
        ver_str_line = handle.read()

    ver_pattern = r'^__version__ = [\'"]([^\'"]*)[\'"]'
    match = re.search(ver_pattern, ver_str_line, re.M)
    if match:
        ver_str = match.group(1)
    else:
        msg = 'Unable to find version string in "{}"'.format(ver_file)
        raise RuntimeError(msg)

    return ver_str


def get_long_description():

    readme_file = 'README.md'
    with open(readme_file, encoding='utf-8') as handle:
        contents = handle.read()

    return contents


setup(
    name='bravais',
    version=get_version(),
    description=('A simple package for representing Bravais lattices. '
                 'Primarily useful to check the passed parameters represent a '
                 'valid Bravais lattice.'),
    long_description=get_long_description(),
    long_description_content_type='text/markdown',
    author='Adam J. Plowman, Maria S. Yankova',
    packages=find_packages(),
    package_data={
        'bravais': ['valid_parameters.yml'],
    },
    install_requires=[
        'numpy',
        'pyyaml',
    ],
    project_urls={
        'Github': 'https://github.com/aplowman/bravais',
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Topic :: Scientific/Engineering :: Physics',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
    ]
)
